from __future__ import annotations

import time
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Iterable

from intermine314.constants import (
    DEFAULT_KEYSET_BATCH_SIZE,
    DEFAULT_LIST_CHUNK_SIZE,
    DEFAULT_TARGETED_EXPORT_PAGE_SIZE,
    DEFAULT_TARGETED_LIST_DESCRIPTION,
    DEFAULT_TARGETED_LIST_NAME_PREFIX,
    DEFAULT_TARGETED_LIST_TAGS,
)
from intermine314.optional_deps import (
    optional_duckdb as _optional_duckdb,
    optional_polars as _optional_polars,
    quote_sql_string as _duckdb_quote,
    require_polars as _require_polars,
)
from intermine314.query_export import write_single_parquet_from_parts


@dataclass(frozen=True)
class TargetedTableSpec:
    name: str
    root_class: str
    views: list[str]
    joins: list[str]
    template_names: list[str]
    template_keywords: list[str]


@dataclass(frozen=True)
class _PreparedTableSpec:
    table: TargetedTableSpec
    template_name: str | None
    template: Any | None
    valid_views: list[str]
    valid_joins: list[str]


def default_oakmine_targeted_tables() -> list[TargetedTableSpec]:
    return [
        TargetedTableSpec(
            name="core_protein",
            root_class="Protein",
            views=[
                "Protein.primaryIdentifier",
                "Protein.primaryAccession",
                "Protein.proteinDescription",
                "Protein.length",
                "Protein.molecularWeight",
                "Protein.EC",
                "Protein.symbol",
            ],
            joins=[],
            template_names=[],
            template_keywords=["protein", "core"],
        ),
        TargetedTableSpec(
            name="edge_go",
            root_class="Protein",
            views=[
                "Protein.primaryIdentifier",
                "Protein.GOTerms.primaryIdentifier",
            ],
            joins=[
                "Protein.GOTerms",
            ],
            template_names=[],
            template_keywords=["go", "gene ontology", "protein"],
        ),
        TargetedTableSpec(
            name="edge_domain",
            root_class="Protein",
            views=[
                "Protein.primaryIdentifier",
                "Protein.domainMotifs.interPro",
                "Protein.domainMotifs.crossRef",
            ],
            joins=[
                "Protein.domainMotifs",
            ],
            template_names=[],
            template_keywords=["interpro", "domain", "protein"],
        ),
    ]


def rank_template_names(names: Iterable[str], keywords: Iterable[str]) -> list[str]:
    lowered_keywords = [str(x).strip().lower() for x in keywords if str(x).strip()]
    if not lowered_keywords:
        return sorted({str(name) for name in names})
    scored: list[tuple[int, str]] = []
    for name in names:
        text = str(name)
        lower = text.lower()
        score = sum(1 for kw in lowered_keywords if kw in lower)
        if score > 0:
            scored.append((score, text))
    scored.sort(key=lambda item: (-item[0], item[1].lower()))
    return [name for _, name in scored]


def _coerce_table_specs(table_specs: list[dict[str, Any]] | None) -> list[TargetedTableSpec]:
    if not table_specs:
        return default_oakmine_targeted_tables()
    parsed: list[TargetedTableSpec] = []
    for raw in table_specs:
        name = str(raw.get("name", "")).strip()
        root_class = str(raw.get("root_class", "Protein")).strip() or "Protein"
        views = [str(v).strip() for v in raw.get("views", []) if str(v).strip()]
        joins = [str(v).strip() for v in raw.get("joins", []) if str(v).strip()]
        template_names = [str(v).strip() for v in raw.get("template_names", []) if str(v).strip()]
        template_keywords = [str(v).strip() for v in raw.get("template_keywords", []) if str(v).strip()]
        if not name or not views:
            continue
        parsed.append(
            TargetedTableSpec(
                name=name,
                root_class=root_class,
                views=views,
                joins=joins,
                template_names=template_names,
                template_keywords=template_keywords,
            )
        )
    return parsed


def _count_parquet_rows(path: Path) -> int | None:
    polars_module = _optional_polars()
    if polars_module is None:
        return None
    try:
        return int(polars_module.scan_parquet(str(path)).select(polars_module.len()).collect().item(0, 0))
    except Exception:
        return None


def _supported_paths(service: Any, root_class: str, paths: list[str]) -> list[str]:
    query = service.new_query(root_class)
    supported: list[str] = []
    for path in paths:
        try:
            _ = query.model.make_path(path, query.get_subclass_dict())
            supported.append(path)
        except Exception:
            continue
    return supported


def _select_template_for_table(service: Any, table: TargetedTableSpec, catalog: list[str]) -> str | None:
    template_set = set(service.templates.keys())
    for name in table.template_names:
        if name in template_set:
            return name
    ranked = rank_template_names(catalog, table.template_keywords)
    return ranked[0] if ranked else None


def _template_constraint_for_list(template: Any, root_class: str, list_name: str) -> dict[str, dict[str, str]] | None:
    for con in template.editable_constraints:
        code = str(getattr(con, "code", "")).strip()
        path = str(getattr(con, "path", "")).strip()
        op = str(getattr(con, "op", "")).strip().upper()
        if not code or not path:
            continue
        if path != root_class and not path.startswith(root_class + "."):
            continue
        if op in {"IN", "ONE OF", "LOOKUP", "="}:
            values = {"value": list_name}
            if op != "IN":
                values["op"] = "IN"
            return {code: values}
    return None


def _export_template_rows_to_parquet(
    template: Any,
    constraints: dict[str, dict[str, str]],
    out_path: Path,
    page_size: int,
) -> tuple[int, float]:
    polars_module = _require_polars("template parquet exports")
    start = 0
    total_rows = 0
    part = 0
    t0 = time.perf_counter()
    with TemporaryDirectory() as tmp:
        staged_dir = Path(tmp)
        while True:
            rows = list(template.results(row="dict", start=start, size=page_size, **constraints))
            if not rows:
                break
            row_count = len(rows)
            polars_module.from_dicts(rows).write_parquet(
                str(staged_dir / f"part-{part:05d}.parquet"),
                compression="zstd",
            )
            total_rows += row_count
            start += row_count
            part += 1
            if row_count < page_size:
                break

        write_single_parquet_from_parts(
            staged_dir=staged_dir,
            target=out_path,
            compression="zstd",
            polars_module=polars_module,
            duckdb_module=_optional_duckdb(),
            duckdb_quote=_duckdb_quote,
        )

    return total_rows, time.perf_counter() - t0


def _prepare_table_plans(
    service: Any,
    specs: list[TargetedTableSpec],
    template_catalog: list[str],
    use_templates_first: bool,
    table_report: dict[str, Any],
) -> list[_PreparedTableSpec]:
    prepared: list[_PreparedTableSpec] = []
    for table in specs:
        ranked_templates = rank_template_names(template_catalog, table.template_keywords)
        table_report[table.name]["template_candidates"].extend(ranked_templates)

        template_name = None
        template = None
        if use_templates_first:
            selected = _select_template_for_table(service, table, template_catalog)
            if selected:
                try:
                    template = service.get_template(selected)
                    template_name = selected
                except Exception as exc:
                    table_report[table.name]["errors"].append(f"template:{selected}:prepare:{exc}")

        valid_views = _supported_paths(service, table.root_class, table.views)
        valid_joins = _supported_paths(service, table.root_class, table.joins)
        if not valid_views:
            table_report[table.name]["errors"].append("fallback_query:prepare:no_supported_views")

        prepared.append(
            _PreparedTableSpec(
                table=table,
                template_name=template_name,
                template=template,
                valid_views=valid_views,
                valid_joins=valid_joins,
            )
        )
    return prepared


def _normalize_identifier_token(row: Any) -> str | None:
    value = row[0] if isinstance(row, (list, tuple)) and row else row
    if value is None:
        return None
    token = str(value).strip()
    return token or None


def _iter_identifier_chunks(
    id_iter: Iterable[Any],
    *,
    list_chunk_size: int,
    row_limit: int | None,
) -> Iterable[tuple[list[str], int]]:
    chunk: list[str] = []
    identifier_count = 0
    for row in id_iter:
        token = _normalize_identifier_token(row)
        if token is None:
            continue
        chunk.append(token)
        identifier_count += 1

        if len(chunk) >= list_chunk_size:
            yield chunk, identifier_count
            chunk = []
        if row_limit is not None and identifier_count >= row_limit:
            break

    if chunk:
        yield chunk, identifier_count


def _export_table_chunk(
    *,
    service: Any,
    prepared: _PreparedTableSpec,
    chunk_index: int,
    list_name: str,
    output_root: Path,
    page_size: int,
    max_workers: int | None,
    ordered: str | bool,
    prefetch: int | None,
    inflight_limit: int | None,
    profile: str,
    large_query_mode: bool,
    keyset_batch_size: int,
    table_report: dict[str, Any],
) -> dict[str, Any] | None:
    table = prepared.table
    table_dir = output_root / table.name
    table_dir.mkdir(parents=True, exist_ok=True)
    out_path = table_dir / f"chunk-{chunk_index:05d}.parquet"

    used_template = None
    wrote_rows = None
    elapsed = None
    valid_views: list[str] = []
    valid_joins: list[str] = []

    if prepared.template is not None:
        try:
            con_values = _template_constraint_for_list(prepared.template, table.root_class, list_name)
            if con_values is not None:
                wrote_rows, elapsed = _export_template_rows_to_parquet(
                    template=prepared.template,
                    constraints=con_values,
                    out_path=out_path,
                    page_size=page_size,
                )
                used_template = prepared.template_name
        except Exception as exc:
            table_report["errors"].append(
                f"template:{prepared.template_name}:chunk:{chunk_index}:{exc}"
            )

    if used_template is None:
        valid_views = list(prepared.valid_views)
        valid_joins = list(prepared.valid_joins)
        if not valid_views:
            table_report["errors"].append(
                f"fallback_query:chunk:{chunk_index}:no_supported_views"
            )
            return None

        t0 = time.perf_counter()
        query = service.new_query(table.root_class)
        query.add_view(*valid_views)
        for join in valid_joins:
            query.add_join(join, "OUTER")
        query.add_constraint(table.root_class, "IN", list_name)
        query.to_parquet(
            str(out_path),
            single_file=True,
            parallel=True,
            page_size=page_size,
            max_workers=max_workers,
            ordered=ordered,
            prefetch=prefetch,
            inflight_limit=inflight_limit,
            profile=profile,
            large_query_mode=large_query_mode,
            pagination="auto",
            keyset_batch_size=keyset_batch_size,
        )
        elapsed = time.perf_counter() - t0
        wrote_rows = _count_parquet_rows(out_path)

    return {
        "chunk_index": chunk_index,
        "list_name": list_name,
        "path": str(out_path),
        "bytes": out_path.stat().st_size if out_path.exists() else 0,
        "rows": wrote_rows,
        "seconds": elapsed,
        "used_template": used_template,
        "fallback_views": valid_views if used_template is None else None,
        "fallback_joins": valid_joins if used_template is None else None,
    }


def export_targeted_tables_with_lists(
    *,
    service: Any,
    root_class: str,
    identifier_path: str,
    output_dir: str | Path,
    table_specs: list[dict[str, Any]] | None = None,
    id_limit: int | None = None,
    list_chunk_size: int = DEFAULT_LIST_CHUNK_SIZE,
    page_size: int = DEFAULT_TARGETED_EXPORT_PAGE_SIZE,
    max_workers: int | None = None,
    ordered: str | bool = "window",
    profile: str = "large_query",
    large_query_mode: bool = True,
    prefetch: int | None = None,
    inflight_limit: int | None = None,
    keyset_batch_size: int = DEFAULT_KEYSET_BATCH_SIZE,
    sleep_seconds: float = 0.0,
    template_keywords: list[str] | None = None,
    template_limit: int = 40,
    use_templates_first: bool = True,
    list_type: str | None = None,
    list_name_prefix: str = DEFAULT_TARGETED_LIST_NAME_PREFIX,
    list_description: str = DEFAULT_TARGETED_LIST_DESCRIPTION,
    list_tags: list[str] | None = None,
) -> dict[str, Any]:
    """
    Export large datasets as a core table plus separate edge tables using list chunks.
    """
    if list_chunk_size <= 0:
        raise ValueError("list_chunk_size must be a positive integer")
    if page_size <= 0:
        raise ValueError("page_size must be a positive integer")

    output_root = Path(output_dir)
    output_root.mkdir(parents=True, exist_ok=True)

    specs = _coerce_table_specs(table_specs)
    if not specs:
        raise ValueError("No valid table specs were supplied")

    list_type = str(list_type or root_class).strip()
    list_name_prefix = str(list_name_prefix).strip() or DEFAULT_TARGETED_LIST_NAME_PREFIX
    list_description = str(list_description).strip() or DEFAULT_TARGETED_LIST_DESCRIPTION
    effective_list_tags = list(list_tags) if list_tags is not None else list(DEFAULT_TARGETED_LIST_TAGS)
    table_report: dict[str, Any] = {
        table.name: {"chunks": [], "errors": [], "template_candidates": []} for table in specs
    }

    template_catalog = list(service.list_templates(include=template_keywords or [], limit=template_limit))
    table_plans = _prepare_table_plans(
        service=service,
        specs=specs,
        template_catalog=template_catalog,
        use_templates_first=use_templates_first,
        table_report=table_report,
    )

    id_query = service.new_query(root_class)
    id_query.add_view(identifier_path)
    row_limit = None if id_limit is None else max(0, int(id_limit))
    id_iter = id_query.run_parallel(
        row="list",
        start=0,
        size=row_limit,
        page_size=page_size,
        max_workers=max_workers,
        ordered=ordered,
        prefetch=prefetch,
        inflight_limit=inflight_limit,
        profile=profile,
        large_query_mode=large_query_mode,
        pagination="auto",
        keyset_path=identifier_path,
        keyset_batch_size=keyset_batch_size,
    )

    created_list_names: list[str] = []
    cleanup_error = None
    chunk_count = 0
    identifier_count = 0
    for chunk_identifiers, identifier_count in _iter_identifier_chunks(
        id_iter,
        list_chunk_size=list_chunk_size,
        row_limit=row_limit,
    ):
        chunk_count += 1
        im_list = None
        try:
            created_lists = service.create_batched_lists(
                chunk_identifiers,
                list_type=list_type,
                chunk_size=list_chunk_size,
                name_prefix=list_name_prefix,
                description=list_description,
                tags=effective_list_tags,
            )
            if not created_lists:
                continue
            im_list = created_lists[0]
            created_list_names.append(im_list.name)

            for prepared in table_plans:
                chunk_report = _export_table_chunk(
                    service=service,
                    prepared=prepared,
                    chunk_index=chunk_count,
                    list_name=im_list.name,
                    output_root=output_root,
                    page_size=page_size,
                    max_workers=max_workers,
                    ordered=ordered,
                    prefetch=prefetch,
                    inflight_limit=inflight_limit,
                    profile=profile,
                    large_query_mode=large_query_mode,
                    keyset_batch_size=keyset_batch_size,
                    table_report=table_report[prepared.table.name],
                )
                if chunk_report is not None:
                    table_report[prepared.table.name]["chunks"].append(chunk_report)
        finally:
            if im_list is not None:
                try:
                    service.delete_lists([im_list.name])
                except Exception as exc:  # pragma: no cover - cleanup best effort
                    if cleanup_error is None:
                        cleanup_error = str(exc)
                    else:
                        cleanup_error = cleanup_error + "; " + str(exc)
        if sleep_seconds > 0:
            time.sleep(sleep_seconds)

    totals = {}
    for table in specs:
        chunks = table_report[table.name]["chunks"]
        totals[table.name] = {
            "files": len(chunks),
            "rows": sum(chunk["rows"] for chunk in chunks if isinstance(chunk.get("rows"), int)),
            "bytes": sum(int(chunk.get("bytes", 0)) for chunk in chunks),
            "seconds": sum(float(chunk.get("seconds", 0.0) or 0.0) for chunk in chunks),
            "templates_used": sum(1 for chunk in chunks if chunk.get("used_template")),
        }

    return {
        "strategy": "targeted_list_exports",
        "root_class": root_class,
        "identifier_path": identifier_path,
        "identifier_count": identifier_count,
        "chunk_size": list_chunk_size,
        "chunk_count": chunk_count if identifier_count else 0,
        "created_lists": created_list_names,
        "cleanup_error": cleanup_error,
        "template_catalog_size": len(template_catalog),
        "tables": table_report,
        "totals": totals,
        "output_dir": str(output_root),
    }



import json
import time
import weakref

from intermine314.constants import DEFAULT_ID_RESOLUTION_MAX_BACKOFF_SECONDS


def get_json(service, path, key):
    text = service.opener.read(service.root + path)
    data = json.loads(text)
    if data["error"] is not None:
        raise Exception(data["error"])
    if key not in data:
        raise Exception(key + " not returned from " + path)
    return data[key]


ONE_MINUTE = DEFAULT_ID_RESOLUTION_MAX_BACKOFF_SECONDS

COMPLETED = set(["SUCCESS", "ERROR"])


class Job(object):
    """
    A Representation of an Identifier Resolution Job
    ================================================

    Users can submit requests to resolve sets of IDs to
    objects in the data-store. These jobs begin in a PENDING
    state, and transition through RUNNING to either SUCCESS
    or ERROR.

    Upon completion, the results of the job may be fetched, and the
    job may be deleted on the server.
    """

    INITIAL_DECAY = 1.25
    INITIAL_BACKOFF = 0.05
    MAX_BACKOFF = ONE_MINUTE

    def __init__(self, service, uid):
        self.service = weakref.proxy(service)
        self.uid = uid
        self.status = None
        self.backoff = Job.INITIAL_BACKOFF
        self.decay = Job.INITIAL_DECAY
        self.max_backoff = Job.MAX_BACKOFF
        if self.uid is None:
            raise Exception("No uid found")

    def poll(self):
        """
        Check to see if the job has been completed, updating the
        status of the job in the process.

        @return: Boolean Whether or not the job is complete.
        """
        if self.status not in COMPLETED:
            backoff = self.backoff
            self.backoff = min(self.max_backoff, backoff * self.decay)
            time.sleep(backoff)
            self.status = self.fetch_status()
        return self.status in COMPLETED

    def fetch_status(self):
        """
        Retrieve the results of this completed job from the server.

        @rtype: dict
        """
        return get_json(self.service, "/ids/{0}/status".format(self.uid), "status")

    def delete(self):
        """
        Delete the job from the server.

        The job should not be used again once this method has been invoked.
        """
        path = "/ids/" + self.uid
        response = self.service.opener.delete(self.service.root + path)
        response_data = json.loads(response)
        if response_data["error"] is not None:
            raise Exception(response_data["error"])

    def fetch_results(self):
        """
        Retrieve the current status of this job from the server.

        @rtype String
        """
        return get_json(self.service, "/ids/{0}/result".format(self.uid), "results")



"""Shared constants for intermine314 runtime defaults."""

DEFAULT_PARALLEL_WORKERS = 16
DEFAULT_PARALLEL_PAGE_SIZE = 1000
DEFAULT_BATCH_SIZE = 5000
DEFAULT_KEYSET_BATCH_SIZE = 2000
DEFAULT_KEYSET_AUTO_MIN_SIZE = 50_000
DEFAULT_PARALLEL_MAX_BUFFERED_ROWS = 100_000

DEFAULT_QUERY_THREAD_NAME_PREFIX = "intermine314-query"
DEFAULT_EXPORT_BATCH_SIZE = 10_000

DEFAULT_LIST_CHUNK_SIZE = 10_000
DEFAULT_LIST_ENTRIES_BATCH_SIZE = DEFAULT_BATCH_SIZE
DEFAULT_TARGETED_EXPORT_PAGE_SIZE = 5_000
DEFAULT_TARGETED_LIST_NAME_PREFIX = "intermine314_targeted_chunk"
DEFAULT_TARGETED_LIST_DESCRIPTION = "Temporary chunk list for targeted benchmark export"
DEFAULT_TARGETED_LIST_TAGS = ("intermine314", "benchmark", "targeted")
DEFAULT_REQUEST_TIMEOUT_SECONDS = 60
DEFAULT_ID_RESOLUTION_MAX_BACKOFF_SECONDS = 60

# Production profile tiers used by mine-aware fetch defaults.
DEFAULT_WORKERS_TIER = 4
SERVER_LIMITED_WORKERS_TIER = 8
FULL_WORKERS_TIER = 16
DEFAULT_PRODUCTION_PROFILE_SWITCH_ROWS = 50_000

PRODUCTION_WORKFLOW_ELT = "elt"
PRODUCTION_WORKFLOW_ETL = "etl"
PRODUCTION_WORKFLOWS = (PRODUCTION_WORKFLOW_ELT, PRODUCTION_WORKFLOW_ETL)

PRODUCTION_PROFILE_ELT_DEFAULT = "elt_default_w4"
PRODUCTION_PROFILE_ELT_SERVER_LIMITED = "elt_server_limited_w8"
PRODUCTION_PROFILE_ELT_FULL = "elt_full_w16"
PRODUCTION_PROFILE_ETL_DEFAULT = "etl_default_w4"
PRODUCTION_PROFILE_ETL_SERVER_LIMITED = "etl_server_limited_w8"
PRODUCTION_PROFILE_ETL_FULL = "etl_full_w16"



import re

PATTERN_STR = "^(?:\\w+\\.)*\\w+$"
PATH_PATTERN = re.compile(PATTERN_STR)


class PathFeature(object):
    def __init__(self, path):
        if path is None:
            raise ValueError("path must not be None")
        try:
            path = path.name
        except Exception:
            pass
        if not PATH_PATTERN.match(path):
            raise TypeError("Path '" + path + "' does not match expected pattern" + PATTERN_STR)
        self.path = path

    def __repr__(self):
        return "<" + self.__class__.__name__ + ": " + self.to_string() + ">"

    def to_string(self):
        return str(self.path)

    def to_dict(self):
        return {"path": self.path}

    @property
    def child_type(self):
        raise AttributeError()


class Join(PathFeature):
    valid_join_styles = ["OUTER", "INNER"]
    INNER = "INNER"
    OUTER = "OUTER"
    child_type = "join"

    def __init__(self, path, style="OUTER"):
        if style.upper() not in Join.valid_join_styles:
            raise TypeError("Unknown join style: " + style)
        self.style = style.upper()
        super(Join, self).__init__(path)

    def to_dict(self):
        d = super(Join, self).to_dict()
        d.update(style=self.style)
        return d

    def __repr__(self):
        return "<" + self.__class__.__name__ + " ".join([":", self.path, self.style]) + ">"


class PathDescription(PathFeature):
    child_type = "pathDescription"

    def __init__(self, path, description):
        self.description = description
        super(PathDescription, self).__init__(path)

    def to_dict(self):
        d = super(PathDescription, self).to_dict()
        d.update(description=self.description)
        return d


class SortOrder(PathFeature):
    ASC = "asc"
    DESC = "desc"
    DIRECTIONS = frozenset(["asc", "desc"])

    def __init__(self, path, order):
        try:
            order = order.lower()
        except Exception:
            pass

        if order not in self.DIRECTIONS:
            raise TypeError("Order must be one of " + str(self.DIRECTIONS) + " - not " + order)
        self.order = order
        super(SortOrder, self).__init__(path)

    def __str__(self):
        return self.path + " " + self.order

    def to_string(self):
        return str(self)


class SortOrderList(object):
    """
    A container implementation for holding sort orders
    ==================================================

    This class exists to hold the sort order information for a
    query. It handles appending elements, and the stringification
    of the sort order.
    """

    def __init__(self, *sos):
        self.sort_orders = []
        self.append(*sos)

    def append(self, *sos):
        """
        Add sort order elements to the sort order list.
        ===============================================

        Elements can be provided as a SortOrder object or
        as a tuple of arguments (path, direction).
        """
        for so in sos:
            if isinstance(so, SortOrder):
                self.sort_orders.append(so)
            elif isinstance(so, tuple):
                self.sort_orders.append(SortOrder(*so))
            else:
                raise TypeError(
                    "Sort orders must be either SortOrder instances," + " or tuples of arguments: I got:" + so + sos
                )

    def __repr__(self):
        return "<" + self.__class__.__name__ + ": [" + str(self) + "]>"

    def __str__(self):
        return " ".join(map(str, self.sort_orders))

    def clear(self):
        self.sort_orders = []

    def is_empty(self):
        return len(self.sort_orders) == 0

    def __len__(self):
        return len(self.sort_orders)

    def __next__(self):
        return next(iter(self.sort_orders))

    def next(self):
        return self.__next__()

    def __iter__(self):
        return iter(self.sort_orders)



import base64
import json
import logging
import re
import sys
from collections import UserDict
from contextlib import closing
from itertools import groupby
from urllib.error import HTTPError
from urllib.parse import urlencode, urlparse
from urllib.request import Request, urlopen
try:
    import requests
except ImportError:  # pragma: no cover - requests is a declared dependency
    requests = None
try:
    import orjson
except ImportError:  # pragma: no cover - optional acceleration
    orjson = None

from intermine314.errors import WebserviceError
from intermine314.model import Attribute, Reference, Collection
from intermine314.constants import DEFAULT_REQUEST_TIMEOUT_SECONDS

from intermine314 import VERSION


def _json_loads(payload):
    if orjson is not None:
        if isinstance(payload, str):
            payload = payload.encode("utf-8")
        return orjson.loads(payload)
    if isinstance(payload, (bytes, bytearray)):
        payload = payload.decode("utf-8")
    return json.loads(payload)


def _json_dumps(payload):
    if orjson is not None:
        return orjson.dumps(payload).decode("utf-8")
    return json.dumps(payload)


class _ResponseBodyAdapter(object):
    def __init__(self, response):
        self._response = response

    def read(self):
        return self._response.content

    def close(self):
        self._response.close()


class _ResponseStreamAdapter(object):
    def __init__(self, response):
        self._response = response
        self._iter = response.iter_lines(decode_unicode=False)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._iter)

    def read(self, size=-1):
        if size is None or size < 0:
            return self._response.content
        return self._response.raw.read(size)

    def close(self):
        self._response.close()


class EnrichmentLine(UserDict):
    """
    An object that represents a result returned from the enrichment service.
    ========================================================================

    These objects operate as dictionaries as well as objects with predefined
    properties.
    """

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return "EnrichmentLine(%s)" % self.data

    def __getattr__(self, name):
        if name is not None:
            key_name = name.replace("_", "-")
            if key_name in self.data:
                return self.data[key_name]
        raise AttributeError(name)


class ResultObject(object):
    """
    An object used to represent result records as returned in jsonobjects format
    ============================================================================

    These objects are backed by a row of data and the class descriptor that
    describes the object. They allow access in standard object style:

        >>> for gene in query.results():
        ...    print(gene.symbol)
        ...    print(map(lambda x: x.name, gene.pathways))

    All objects will have "id" and "type" properties. The type refers to the
    actual type of this object: if it is a subclass of the one requested, the
    subclass name will be returned. The "id" refers to the internal database id
    of the object, and is a guarantor of object identity.

    """

    def __init__(self, data, cld, view=None):
        if view is None:
            view = []
        stripped = [v[v.find(".") + 1 :] for v in view]
        self.selected_attributes = [v for v in stripped if "." not in v]
        self.reference_paths = dict(((k, list(i)) for k, i in groupby(stripped, lambda x: x[: x.find(".") + 1])))
        self._data = data
        # Make sure this object has the most specific class desc. possible
        class_name = data["class"]
        if "class" not in data or cld.name == class_name:
            self._cld = cld
        else:  # this could be a composed class - behave accordingly.
            self._cld = cld.model.get_class(class_name)

        self._attr_cache = {}

    def __str__(self):
        dont_show = set(["objectId", "class"])
        return "%s(%s)" % (
            self._cld.name,
            ",  ".join(
                "%s = %r" % (k, v)
                for k, v in list(self._data.items())
                if not isinstance(v, dict) and not isinstance(v, list) and k not in dont_show
            ),
        )

    def __repr__(self):
        dont_show = set(["objectId", "class"])
        return "%s(%s)" % (
            self._cld.name,
            ", ".join("%s = %r" % (k, getattr(self, k)) for k in self._data if k not in dont_show),
        )

    def __getattr__(self, name):
        if name in self._attr_cache:
            return self._attr_cache[name]

        if name == "type":
            return self._data["class"]

        fld = self._cld.get_field(name)
        attr = None
        if isinstance(fld, Attribute):
            if name in self._data:
                attr = self._data[name]
            if attr is None:
                attr = self._fetch_attr(fld)
        elif isinstance(fld, Reference):
            ref_paths = self._get_ref_paths(fld)
            if name in self._data:
                data = self._data[name]
            else:
                data = self._fetch_reference(fld)
            if isinstance(fld, Collection):
                if data is None:
                    attr = []
                else:
                    attr = [ResultObject(x, fld.type_class, ref_paths) for x in data]
            else:
                if data is None:
                    attr = None
                else:
                    attr = ResultObject(data, fld.type_class, ref_paths)
        else:
            raise WebserviceError("Inconsistent model - This should never happen")
        self._attr_cache[name] = attr
        return attr

    def _get_ref_paths(self, fld):
        if fld.name + "." in self.reference_paths:
            return self.reference_paths[fld.name + "."]
        else:
            return []

    @property
    def id(self):
        """Return the internal DB identifier of this object. Or None if this is not an InterMine object"""
        return self._data.get("objectId")

    def _fetch_attr(self, fld):
        if fld.name in self.selected_attributes:
            return None  # Was originally selected - no point asking twice
        c = self._cld
        if "id" not in c:
            return None  # Cannot reliably fetch anything without access to the objectId.
        q = c.model.service.query(c, fld).where(id=self.id)
        r = q.first()
        return r._data[fld.name] if fld.name in r._data else None

    def _fetch_reference(self, ref):
        if ref.name + "." in self.reference_paths:
            return None  # Was originally selected - no point asking twice.
        c = self._cld
        if "id" not in c:
            return None  # Cannot reliably fetch anything without access to the objectId.
        q = c.model.service.query(ref).outerjoin(ref).where(id=self.id)
        r = q.first()
        return r._data[ref.name] if ref.name in r._data else None


class ResultRow(object):
    """
    An object for representing a row of data received back from the server.
    =======================================================================

    ResultRows provide access to the fields of the row through index lookup. However,
    for convenience both list indexes and dictionary keys can be used. So the
    following all work:

        >>> # Assuming the view is "Gene.symbol", "Gene.organism.name":
        >>> row[0] == row["symbol"] == row["Gene.symbol"] == row(0) == row("symbol")
        ... True

    """

    def __init__(self, data, views):
        self.data = data
        self.views = views
        self.index_map = None

    def __len__(self):
        """Return the number of cells in this row"""
        return len(self.data)

    def __iter__(self):
        """Return the list view of the row, so each cell can be processed"""
        return iter(self.to_l())

    def _get_index_for(self, key):
        if self.index_map is None:
            self.index_map = {}
            for i in range(len(self.views)):
                view = self.views[i]
                headless_view = re.sub("^[^.]+.", "", view)
                self.index_map[view] = i
                self.index_map[headless_view] = i

        return self.index_map[key]

    def __str__(self):
        root = re.sub(r"\..*$", "", self.views[0])
        parts = [root + ":"]
        for view in self.views:
            short_form = re.sub("^[^.]+.", "", view)
            value = self[view]
            parts.append(short_form + "=" + repr(value))
        return " ".join(parts)

    def __call__(self, name):
        return self.__getitem__(name)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.data[key]
        elif isinstance(key, slice):
            return self.data[key]
        else:
            index = self._get_index_for(key)
            return self.data[index]

    def to_l(self):
        """Return a list view of this row"""
        return list(self.data)

    def to_d(self):
        """Return a dictionary view of this row"""
        return dict(self.items())

    def items(self):
        return [(view, self[view]) for view in self.views]

    def keys(self):
        return list(self.views)

    def values(self):
        return self.to_l()


class TableResultRow(ResultRow):
    """
    A class for parsing results from the jsonrows data format.
    """

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.data[key]["value"]
        elif isinstance(key, slice):
            return [x["value"] for x in self.data[key]]
        else:
            index = self._get_index_for(key)
            return self.data[index]["value"]

    def to_l(self):
        """Return a list view of this row"""
        return [x["value"] for x in self.data]


def encode_str(value):
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return value
    return str(value)


def decode_binary(value):
    if isinstance(value, (bytes, bytearray)):
        return value.decode("utf-8")
    return value


def encode_dict(input_d):
    return {encode_str(k): encode_str(v) for k, v in input_d.items()}


def _is_blank_query_param_error(exc):
    text = str(exc).lower()
    return "query" in text and "must not be blank" in text


class ResultIterator(object):
    """
    A facade over the internal iterator object
    ==========================================

    These objects handle the iteration over results
    in the formats requested by the user. They are responsible
    for generating an appropriate parser,
    connecting the parser to the results, and delegating
    iteration appropriately.
    """

    PARSED_FORMATS = frozenset(["rr", "list", "dict"])
    STRING_FORMATS = frozenset(["tsv", "csv", "count"])
    JSON_FORMATS = frozenset(["jsonrows", "jsonobjects", "json"])
    ROW_FORMATS = PARSED_FORMATS | STRING_FORMATS | JSON_FORMATS

    def __init__(self, service, path, params, rowformat, view, cld=None):
        """
        Constructor
        ===========

        Services are responsible for getting result iterators. You will
        not need to create one manually.

        @param root: The root path (eg: "https://www.flymine.org/query/service")
        @type root: string
        @param path: The resource path (eg: "/query/results")
        @type path: string
        @param params: The query parameters for this request
        @type params: dict
        @param rowformat: One of "rr", "object", "count", "dict", "list", "tsv", "csv", "jsonrows", "jsonobjects", "json"
        @type rowformat: string
        @param view: The output columns
        @type view: list
        @param opener: A url opener (user-agent)
        @type opener: urllib.URLopener

        @raise ValueError: if the row format is incorrect
        @raise WebserviceError: if the request is unsuccessful
        """
        if rowformat.startswith("object"):  # Accept "object", "objects", "objectformat", etc...
            rowformat = "jsonobjects"  # these are synonymous
        if rowformat not in self.ROW_FORMATS:
            raise ValueError(
                "'%s' is not one of the valid row formats (%s)" % (rowformat, repr(list(self.ROW_FORMATS)))
            )

        self.row = ResultRow if service.version >= 8 else TableResultRow

        if rowformat in self.PARSED_FORMATS:
            if service.version >= 8:
                params.update({"format": "json"})
            else:
                params.update({"format": "jsonrows"})
        elif rowformat == "tsv":
            params.update({"format": "tab"})
        else:
            params.update({"format": rowformat})

        self.url = service.root + path
        self.data = urlencode(encode_dict(params), True)
        self.view = view
        self.opener = service.opener
        self.cld = cld
        self.rowformat = rowformat
        self._it = None

    def __len__(self):
        """
        Return the number of items in this iterator
        ===========================================

        Note that this requires iterating over the full result set, making the
        request in the process.
        """
        c = 0
        for x in self:
            c += 1
        return c

    def __iter__(self):
        """
        Return an iterator over the results
        ===================================

        Returns the internal iterator object.
        """
        try:
            con = self.opener.open(self.url, self.data)
        except WebserviceError as post_error:
            join_char = "&" if "?" in self.url else "?"
            fallback_url = self.url + join_char + self.data
            try:
                con = self.opener.open(fallback_url, method="GET")
            except WebserviceError:
                if _is_blank_query_param_error(post_error):
                    raise
                raise post_error
        identity = lambda x: x
        if self.rowformat in {"tsv", "csv", "count"}:
            return FlatFileIterator(con, identity)
        if self.rowformat in {"json", "jsonrows"}:
            return JSONIterator(con, identity)
        if self.rowformat == "list":
            return JSONIterator(con, lambda x: self.row(x, self.view).to_l())
        if self.rowformat == "rr":
            return JSONIterator(con, lambda x: self.row(x, self.view))
        if self.rowformat == "dict":
            return JSONIterator(con, lambda x: self.row(x, self.view).to_d())
        if self.rowformat == "jsonobjects":
            return JSONIterator(con, lambda x: ResultObject(x, self.cld, self.view))
        raise ValueError("Couldn't get iterator for " + self.rowformat)

    def __next__(self):
        if self._it is None:
            self._it = iter(self)
        try:
            return next(self._it)
        except StopIteration:
            self._it = None
            raise

    def next(self):
        return self.__next__()


class FlatFileIterator(object):
    """
    An iterator for handling results returned as a flat file (TSV/CSV).
    ===================================================================

    This iterator can be used as the sub iterator in a ResultIterator
    """

    def __init__(self, connection, parser):
        """
        Constructor
        ===========

        @param connection: The source of data
        @type connection: socket.socket
        @param parser: a handler for each row of data
        @type parser: Parser
        """
        self.connection = connection
        self.parser = parser

    def __iter__(self):
        return self

    def __next__(self):
        line = decode_binary(next(self.connection)).strip()
        if line.startswith("[ERROR]"):
            raise WebserviceError(line)
        return self.parser(line)

    def next(self):
        return self.__next__()


class JSONIterator(object):
    """
    An iterator for handling results returned in the JSONRows format
    ================================================================

    This iterator can be used as the sub iterator in a ResultIterator
    """

    LOG = logging.getLogger("JSONIterator")

    def __init__(self, connection, parser):
        """
        Constructor
        ===========

        @param connection: The source of data
        @type connection: socket.socket
        @param parser: a handler for each row of data
        @type parser: Parser
        """
        self.connection = connection
        self.parser = parser
        self.header = ""
        self.footer = ""
        self.parse_header()
        self._is_finished = False

    def __iter__(self):
        return self

    def __next__(self):
        if self._is_finished:
            raise StopIteration
        return self.get_next_row_from_connection()

    def next(self):
        return self.__next__()

    def parse_header(self):
        """Reads out the header information from the connection"""
        self.LOG.debug("Connection = {0}".format(self.connection))
        try:
            while True:
                line = decode_binary(next(self.connection)).strip()
                self.header += line
                if line.endswith('"results":['):
                    return
        except StopIteration:
            raise WebserviceError("The connection returned a bad header" + self.header)

    def check_return_status(self):
        """
        Perform status checks
        =====================

        The footer containts information as to whether the result
        set was successfully transferred in its entirety. This
        method makes sure we don't silently accept an
        incomplete result set.

        @raise WebserviceError: if the footer indicates there was an error
        """
        container = self.header + self.footer
        info = None
        try:
            info = _json_loads(container)
        except Exception:
            raise WebserviceError("Error parsing JSON container: " + container)

        if not info["wasSuccessful"]:
            raise WebserviceError(info["statusCode"], info["error"])

    def get_next_row_from_connection(self):
        """
        Reads the connection to get the next row, and sends it to the parser

        @raise WebserviceError: if the connection is interrupted
        """
        next_row = None
        try:
            line = decode_binary(next(self.connection))
            if line.startswith("]"):
                self.footer += line
                for otherline in self.connection:
                    self.footer += decode_binary(otherline)
                self.check_return_status()
            else:
                line = line.strip().strip(",")
                if len(line) > 0:
                    try:
                        row = _json_loads(line)
                    except Exception as e:
                        raise WebserviceError("Error parsing line from results: '" + line + "' - " + str(e))
                    next_row = self.parser(row)
        except StopIteration:
            raise WebserviceError("Connection interrupted")

        if next_row is None:
            self._is_finished = True
            raise StopIteration
        else:
            return next_row


def encode_headers(headers):
    return {encode_header_value(k): encode_header_value(v) for k, v in headers.items()}


def encode_header_value(value):
    if isinstance(value, bytes):
        return value.decode("ascii")
    return str(value)


class InterMineURLOpener(object):
    """
    Specific implementation of FancyURLopener for this client
    ================================================================

    Provides user agent and authentication headers, and handling of errors
    """

    USER_AGENT = "InterMine-Client-{0}/python-{1}".format(VERSION, sys.version_info)
    PLAIN_TEXT = "text/plain"
    JSON = "application/json"
    _logged_urllib_fallback = False

    def __init__(self, credentials=None, token=None, request_timeout=DEFAULT_REQUEST_TIMEOUT_SECONDS):
        """
        Constructor
        ===========

        InterMineURLOpener((username, password)) S{->} InterMineURLOpener

        Return a new url-opener with the appropriate credentials
        """
        self.token = token
        if credentials and len(credentials) == 2:
            encoded = "{0}:{1}".format(*credentials).encode("utf8")
            base64string = "Basic {0}".format(base64.encodebytes(encoded)[:-1].decode("ascii"))
            self.auth_header = base64string
            self.using_authentication = True
        elif self.token is not None:
            token_header = "Token {0}".format(self.token)
            self.auth_header = token_header
            self.using_authentication = True
        else:
            self.using_authentication = False
        self.request_timeout = request_timeout
        self._session = requests.Session() if requests is not None else None

    def clone(self):
        clone = InterMineURLOpener(request_timeout=self.request_timeout)
        clone.token = self.token
        clone.using_authentication = self.using_authentication
        clone._session = self._session
        if self.using_authentication:
            clone.auth_header = self.auth_header
        return clone

    def headers(self, content_type=None, accept=None):
        h = {"User-Agent": self.USER_AGENT}
        if self.using_authentication:
            h["Authorization"] = self.auth_header
        if content_type is not None:
            h["Content-Type"] = content_type
        if accept is not None:
            h["Accept"] = accept
        return h

    def post_plain_text(self, url, body):
        return self.post_content(url, body, InterMineURLOpener.PLAIN_TEXT)

    def post_content(self, url, body, mimetype, charset="utf-8"):
        content_type = "{0}; charset={1}".format(mimetype, charset)

        with closing(self.open(url, body, {"Content-Type": content_type})) as f:
            return f.read()

    def open(self, url, data=None, headers=None, method=None, timeout=None):
        url = self.prepare_url(url)
        effective_timeout = self.request_timeout if timeout is None else timeout
        if data is None:
            buff = None
        elif isinstance(data, (bytes, bytearray)):
            buff = data
        else:
            buff = data.encode("utf-8")
        hs = self.headers()
        if headers is not None:
            hs.update(headers)
        if buff is not None and "Content-Type" not in hs:
            hs["Content-Type"] = "application/x-www-form-urlencoded; charset=utf-8"
        if method is None:
            method = "POST" if buff is not None else "GET"

        if self._session is not None:
            try:
                resp = self._session.request(
                    method,
                    url,
                    data=buff,
                    headers=hs,
                    stream=True,
                    timeout=effective_timeout,
                )
            except requests.RequestException as e:
                raise WebserviceError("Request failed", 0, str(e), str(e))
            if resp.status_code >= 400:
                fp = _ResponseBodyAdapter(resp)
                args = (
                    url,
                    fp,
                    resp.status_code,
                    resp.reason,
                    resp.headers,
                )
                handler = {
                    400: self.http_error_400,
                    401: self.http_error_401,
                    403: self.http_error_403,
                    404: self.http_error_404,
                    500: self.http_error_500,
                }.get(resp.status_code, self.http_error_default)
                handler(*args)
            return _ResponseStreamAdapter(resp)

        if not self._logged_urllib_fallback:
            logging.getLogger("InterMineURLOpener").warning(
                "Falling back to urllib.request (no pooled session client)."
            )
            self.__class__._logged_urllib_fallback = True

        req = Request(url, buff, headers=hs)
        if method is not None:
            req.get_method = lambda: method
        try:
            return urlopen(req, timeout=effective_timeout)
        except HTTPError as e:
            args = (
                url,
                e,
                e.code,
                e.reason,
                e.headers,
            )
            handler = {
                400: self.http_error_400,
                401: self.http_error_401,
                403: self.http_error_403,
                404: self.http_error_404,
                500: self.http_error_500,
            }.get(e.code, self.http_error_default)
            handler(*args)

    def read(self, url, data=None):
        with closing(self.open(url, data)) as conn:
            content = conn.read()
            return decode_binary(content)

    def prepare_url(self, url):
        # Generally unnecessary these days - will be deprecated one of these days.
        if self.token:
            token_param = urlencode(encode_dict(dict(token=self.token)))
            o = urlparse(url)
            if o.query:
                url += "&" + token_param
            else:
                url += "?" + token_param

        return url

    def delete(self, url):
        with closing(self.open(url, method="DELETE")) as f:
            return f.read()

    def http_error_default(self, url, fp, errcode, errmsg, headers):
        """Re-implementation of http_error_default, with content now supplied by default"""
        content = fp.read()
        fp.close()
        raise WebserviceError(errcode, errmsg, content)

    def http_error_400(self, url, fp, errcode, errmsg, headers, data=None):
        """
        Handle 400 HTTP errors, attempting to return informative error messages
        =======================================================================

        400 errors indicate that something about our request was incorrect

        @raise WebserviceError: in all circumstances

        """
        content = fp.read()
        fp.close()
        try:
            message = _json_loads(content)["error"]
        except Exception:
            message = content
        raise WebserviceError("There was a problem with our request", errcode, errmsg, message)

    def http_error_401(self, url, fp, errcode, errmsg, headers, data=None):
        """
        Handle 401 HTTP errors, attempting to return informative error messages
        =======================================================================

        401 errors indicate we don't have sufficient permission for the resource
        we requested - usually a list or a tempate

        @raise WebserviceError: in all circumstances

        """
        content = fp.read()
        fp.close()
        if self.using_authentication:
            auth = self.auth_header
            raise WebserviceError("Insufficient permissions - {0}".format(auth), errcode, errmsg, content)
        else:
            raise WebserviceError("No permissions - not logged in", errcode, errmsg, content)

    def http_error_403(self, url, fp, errcode, errmsg, headers, data=None):
        """
        Handle 403 HTTP errors, attempting to return informative error messages
        =======================================================================

        401 errors indicate we don't have sufficient permission for the resource
        we requested - usually a list or a tempate

        @raise WebserviceError: in all circumstances

        """
        content = fp.read()
        fp.close()
        try:
            message = _json_loads(content)["error"]
        except Exception:
            message = content
        if self.using_authentication:
            raise WebserviceError("Insufficient permissions", errcode, errmsg, message)
        else:
            raise WebserviceError("No permissions - not logged in", errcode, errmsg, message)

    def http_error_404(self, url, fp, errcode, errmsg, headers, data=None):
        """
        Handle 404 HTTP errors, attempting to return informative error messages
        =======================================================================

        404 errors indicate that the requested resource does not exist - usually
        a template that is not longer available.

        @raise WebserviceError: in all circumstances

        """
        content = fp.read()
        fp.close()
        try:
            message = _json_loads(content)["error"]
        except Exception:
            message = content
        raise WebserviceError("Missing resource", errcode, errmsg, message)

    def http_error_500(self, url, fp, errcode, errmsg, headers, data=None):
        """
        Handle 500 HTTP errors, attempting to return informative error messages
        =======================================================================

        500 errors indicate that the server borked during the request - ie: it wasn't
        our fault.

        @raise WebserviceError: in all circumstances

        """
        content = fp.read()
        fp.close()
        try:
            message = _json_loads(content)["error"]
        except Exception:
            message = content
        raise WebserviceError("Internal server error", errcode, errmsg, message)



from intermine314.util import ReadableException


class UnimplementedError(Exception):
    pass


class ServiceError(ReadableException):
    """Errors in the creation and use of the Service object"""

    pass


class WebserviceError(IOError):
    """Errors from interaction with the webservice"""

    pass



from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from urllib.parse import urlparse

from intermine314.constants import (
    DEFAULT_PARALLEL_WORKERS,
    DEFAULT_PRODUCTION_PROFILE_SWITCH_ROWS,
    DEFAULT_WORKERS_TIER,
    FULL_WORKERS_TIER,
    PRODUCTION_PROFILE_ELT_DEFAULT,
    PRODUCTION_PROFILE_ELT_FULL,
    PRODUCTION_PROFILE_ELT_SERVER_LIMITED,
    PRODUCTION_PROFILE_ETL_DEFAULT,
    PRODUCTION_PROFILE_ETL_FULL,
    PRODUCTION_PROFILE_ETL_SERVER_LIMITED,
    PRODUCTION_WORKFLOW_ELT,
    PRODUCTION_WORKFLOW_ETL,
    PRODUCTION_WORKFLOWS,
    SERVER_LIMITED_WORKERS_TIER,
)

try:
    import tomllib
except Exception:  # pragma: no cover - Python 3.14 includes tomllib
    tomllib = None


DEFAULT_BENCHMARK_SMALL_PROFILE = "benchmark_profile_3"
DEFAULT_BENCHMARK_LARGE_PROFILE = "benchmark_profile_1"
DEFAULT_BENCHMARK_FALLBACK_PROFILE = DEFAULT_BENCHMARK_SMALL_PROFILE
DEFAULT_PROFILE_SWITCH_ROWS = DEFAULT_PRODUCTION_PROFILE_SWITCH_ROWS

DEFAULT_BENCHMARK_PROFILES = {
    DEFAULT_BENCHMARK_LARGE_PROFILE: {
        "include_legacy_baseline": False,
        "workers": [4, 8, 12, 16],
    },
    "benchmark_profile_2": {
        "include_legacy_baseline": False,
        "workers": [4, 6, 8],
    },
    "benchmark_profile_3": {
        "include_legacy_baseline": True,
        "workers": [4, 8, 12, 16],
    },
    "benchmark_profile_4": {
        "include_legacy_baseline": True,
        "workers": [4, 6, 8],
    },
}

DEFAULT_PRODUCTION_PROFILES = {
    PRODUCTION_PROFILE_ELT_DEFAULT: {
        "workflow": PRODUCTION_WORKFLOW_ELT,
        "workers": DEFAULT_WORKERS_TIER,
        "pipeline": "parquet_duckdb",
        "parallel_profile": "large_query",
        "ordered": "unordered",
        "large_query_mode": True,
    },
    PRODUCTION_PROFILE_ELT_SERVER_LIMITED: {
        "workflow": PRODUCTION_WORKFLOW_ELT,
        "workers": SERVER_LIMITED_WORKERS_TIER,
        "pipeline": "parquet_duckdb",
        "parallel_profile": "large_query",
        "ordered": "unordered",
        "large_query_mode": True,
    },
    PRODUCTION_PROFILE_ELT_FULL: {
        "workflow": PRODUCTION_WORKFLOW_ELT,
        "workers": FULL_WORKERS_TIER,
        "pipeline": "parquet_duckdb",
        "parallel_profile": "large_query",
        "ordered": "unordered",
        "large_query_mode": True,
    },
    PRODUCTION_PROFILE_ETL_DEFAULT: {
        "workflow": PRODUCTION_WORKFLOW_ETL,
        "workers": DEFAULT_WORKERS_TIER,
        "pipeline": "polars_duckdb",
        "parallel_profile": "large_query",
        "ordered": "unordered",
        "large_query_mode": True,
    },
    PRODUCTION_PROFILE_ETL_SERVER_LIMITED: {
        "workflow": PRODUCTION_WORKFLOW_ETL,
        "workers": SERVER_LIMITED_WORKERS_TIER,
        "pipeline": "polars_duckdb",
        "parallel_profile": "large_query",
        "ordered": "unordered",
        "large_query_mode": True,
    },
    PRODUCTION_PROFILE_ETL_FULL: {
        "workflow": PRODUCTION_WORKFLOW_ETL,
        "workers": FULL_WORKERS_TIER,
        "pipeline": "polars_duckdb",
        "parallel_profile": "large_query",
        "ordered": "unordered",
        "large_query_mode": True,
    },
}

DEFAULT_PRODUCTION_PROFILE_BY_WORKFLOW = {
    PRODUCTION_WORKFLOW_ELT: PRODUCTION_PROFILE_ELT_DEFAULT,
    PRODUCTION_WORKFLOW_ETL: PRODUCTION_PROFILE_ETL_DEFAULT,
}

FULL_PRODUCTION_PROFILE_BY_WORKFLOW = {
    PRODUCTION_WORKFLOW_ELT: PRODUCTION_PROFILE_ELT_FULL,
    PRODUCTION_WORKFLOW_ETL: PRODUCTION_PROFILE_ETL_FULL,
}

SERVER_LIMITED_PROFILE_BY_WORKFLOW = {
    PRODUCTION_WORKFLOW_ELT: PRODUCTION_PROFILE_ELT_SERVER_LIMITED,
    PRODUCTION_WORKFLOW_ETL: PRODUCTION_PROFILE_ETL_SERVER_LIMITED,
}


def _to_int_list(values):
    result = []
    for value in values:
        try:
            ivalue = int(value)
        except Exception:
            continue
        if ivalue <= 0:
            continue
        result.append(ivalue)
    return result


def _normalize_workflow(workflow):
    value = str(workflow or PRODUCTION_WORKFLOW_ELT).strip().lower()
    if value not in PRODUCTION_WORKFLOWS:
        choices = ", ".join(PRODUCTION_WORKFLOWS)
        raise ValueError(f"workflow must be one of: {choices}")
    return value


def _workers_to_production_profile(workflow, workers):
    workflow = _normalize_workflow(workflow)
    if int(workers) <= DEFAULT_WORKERS_TIER:
        return DEFAULT_PRODUCTION_PROFILE_BY_WORKFLOW[workflow]
    if int(workers) <= SERVER_LIMITED_WORKERS_TIER:
        return SERVER_LIMITED_PROFILE_BY_WORKFLOW[workflow]
    return FULL_PRODUCTION_PROFILE_BY_WORKFLOW[workflow]


def _standard_mine_profile(
    *,
    display_name,
    host_patterns,
    path_prefixes,
    default_workers=FULL_WORKERS_TIER,
    production_large_workers=FULL_WORKERS_TIER,
    production_switch_rows=DEFAULT_PROFILE_SWITCH_ROWS,
    benchmark_small_profile=DEFAULT_BENCHMARK_SMALL_PROFILE,
    benchmark_large_profile=DEFAULT_BENCHMARK_LARGE_PROFILE,
    benchmark_switch_rows=DEFAULT_PROFILE_SWITCH_ROWS,
    production_elt_small_profile=None,
    production_elt_large_profile=None,
    production_etl_small_profile=None,
    production_etl_large_profile=None,
):
    default_workers = int(default_workers)
    production_large_workers = int(production_large_workers)
    if production_elt_small_profile is None:
        production_elt_small_profile = _workers_to_production_profile(PRODUCTION_WORKFLOW_ELT, default_workers)
    if production_elt_large_profile is None:
        production_elt_large_profile = _workers_to_production_profile(PRODUCTION_WORKFLOW_ELT, production_large_workers)
    if production_etl_small_profile is None:
        production_etl_small_profile = _workers_to_production_profile(PRODUCTION_WORKFLOW_ETL, default_workers)
    if production_etl_large_profile is None:
        production_etl_large_profile = _workers_to_production_profile(PRODUCTION_WORKFLOW_ETL, production_large_workers)

    return {
        "display_name": display_name,
        "host_patterns": host_patterns,
        "path_prefixes": path_prefixes,
        # Legacy worker-resolution fields still used by existing query defaults.
        "default_workers": default_workers,
        "large_query_threshold_rows": int(production_switch_rows),
        "workers_above_threshold": production_large_workers,
        "workers_when_size_unknown": production_large_workers,
        # Production profile mapping (ELT + ETL).
        "production_profile_switch_rows": int(production_switch_rows),
        "production_elt_small_profile": str(production_elt_small_profile),
        "production_elt_large_profile": str(production_elt_large_profile),
        "production_etl_small_profile": str(production_etl_small_profile),
        "production_etl_large_profile": str(production_etl_large_profile),
        # Benchmark profile mapping.
        "benchmark_profile": benchmark_small_profile,
        "benchmark_small_profile": benchmark_small_profile,
        "benchmark_switch_threshold_rows": int(benchmark_switch_rows),
        "benchmark_large_profile": benchmark_large_profile,
    }


DEFAULT_REGISTRY = {
    "legumemine": {
        "display_name": "LegumeMine",
        "host_patterns": ["mines.legumeinfo.org"],
        "path_prefixes": ["/legumemine"],
        "default_workers": DEFAULT_WORKERS_TIER,
        "large_query_threshold_rows": DEFAULT_PROFILE_SWITCH_ROWS,
        "workers_above_threshold": DEFAULT_WORKERS_TIER,
        "workers_when_size_unknown": DEFAULT_WORKERS_TIER,
        "production_profile_switch_rows": DEFAULT_PROFILE_SWITCH_ROWS,
        "production_elt_small_profile": PRODUCTION_PROFILE_ELT_DEFAULT,
        "production_elt_large_profile": PRODUCTION_PROFILE_ELT_DEFAULT,
        "production_etl_small_profile": PRODUCTION_PROFILE_ETL_DEFAULT,
        "production_etl_large_profile": PRODUCTION_PROFILE_ETL_DEFAULT,
        "benchmark_profile": DEFAULT_BENCHMARK_SMALL_PROFILE,
        "benchmark_small_profile": DEFAULT_BENCHMARK_SMALL_PROFILE,
        "benchmark_switch_threshold_rows": DEFAULT_PROFILE_SWITCH_ROWS,
        "benchmark_small_workers": [],
        "benchmark_small_include_legacy_baseline": False,
        "benchmark_large_profile": "benchmark_profile_2",
    },
    "maizemine": _standard_mine_profile(
        display_name="MaizeMine",
        host_patterns=["maizemine.rnet.missouri.edu"],
        path_prefixes=["/maizemine"],
        default_workers=SERVER_LIMITED_WORKERS_TIER,
        production_large_workers=SERVER_LIMITED_WORKERS_TIER,
        benchmark_small_profile="benchmark_profile_4",
        benchmark_large_profile="benchmark_profile_2",
        production_elt_small_profile=PRODUCTION_PROFILE_ELT_SERVER_LIMITED,
        production_elt_large_profile=PRODUCTION_PROFILE_ELT_SERVER_LIMITED,
        production_etl_small_profile=PRODUCTION_PROFILE_ETL_SERVER_LIMITED,
        production_etl_large_profile=PRODUCTION_PROFILE_ETL_SERVER_LIMITED,
    ),
    "thalemine": _standard_mine_profile(
        display_name="ThaleMine",
        host_patterns=["bar.utoronto.ca"],
        path_prefixes=["/thalemine"],
    ),
    "oakmine": _standard_mine_profile(
        display_name="OakMine",
        host_patterns=["urgi.versailles.inra.fr", "urgi.versailles.inrae.fr"],
        path_prefixes=["/OakMine_PM1N"],
    ),
    "wheatmine": _standard_mine_profile(
        display_name="WheatMine",
        host_patterns=["urgi.versailles.inra.fr", "urgi.versailles.inrae.fr"],
        path_prefixes=["/WheatMine"],
    ),
}

_CACHE = None


def _config_path():
    return Path(__file__).resolve().parent.parent / "config" / "mine-parallel-preferences.toml"


def _default_mines_copy():
    return deepcopy(DEFAULT_REGISTRY)


def _default_benchmark_profiles_copy():
    return deepcopy(DEFAULT_BENCHMARK_PROFILES)


def _default_production_profiles_copy():
    return deepcopy(DEFAULT_PRODUCTION_PROFILES)


def _merge_mines(loaded):
    merged = _default_mines_copy()

    defaults_block = loaded.get("defaults", {}) if isinstance(loaded, dict) else {}
    mine_defaults = defaults_block.get("mine", {}) if isinstance(defaults_block, dict) else {}
    if not isinstance(mine_defaults, dict):
        mine_defaults = {}

    raw_mines = loaded.get("mines", {}) if isinstance(loaded, dict) else {}
    if not isinstance(raw_mines, dict):
        raw_mines = {}

    for name, profile in raw_mines.items():
        if not isinstance(profile, dict):
            continue
        base = dict(merged.get(name, {}))
        if mine_defaults:
            base.update(mine_defaults)
        base.update(profile)
        merged[name] = base

    if mine_defaults:
        for name, profile in merged.items():
            if name not in raw_mines:
                base = dict(profile)
                base.update(mine_defaults)
                merged[name] = base
    return merged


def _merge_benchmark_profiles(loaded):
    merged = _default_benchmark_profiles_copy()
    raw_profiles = loaded.get("benchmark_profiles", {}) if isinstance(loaded, dict) else {}
    if not isinstance(raw_profiles, dict):
        return merged

    for name, profile in raw_profiles.items():
        if not isinstance(profile, dict):
            continue
        base = dict(merged.get(name, {}))
        base.update(profile)
        merged[name] = base
    return merged


def _merge_production_profiles(loaded):
    merged = _default_production_profiles_copy()
    raw_profiles = loaded.get("production_profiles", {}) if isinstance(loaded, dict) else {}
    if not isinstance(raw_profiles, dict):
        return merged

    for name, profile in raw_profiles.items():
        if not isinstance(profile, dict):
            continue
        base = dict(merged.get(name, {}))
        base.update(profile)
        merged[name] = base
    return merged


def _load_registry():
    global _CACHE
    if _CACHE is not None:
        return _CACHE

    data = {
        "mines": _default_mines_copy(),
        "benchmark_profiles": _default_benchmark_profiles_copy(),
        "production_profiles": _default_production_profiles_copy(),
    }
    cfg = _config_path()
    if cfg.exists() and tomllib is not None:
        try:
            loaded = tomllib.loads(cfg.read_text(encoding="utf-8"))
            if isinstance(loaded, dict):
                data["mines"] = _merge_mines(loaded)
                data["benchmark_profiles"] = _merge_benchmark_profiles(loaded)
                data["production_profiles"] = _merge_production_profiles(loaded)
        except Exception:
            pass
    _CACHE = data
    return _CACHE


def _normalize_service_root(service_root):
    if not service_root:
        return "", ""
    parsed = urlparse(service_root)
    host = (parsed.hostname or "").lower()
    path = (parsed.path or "").rstrip("/")
    if path.endswith("/service"):
        path = path[: -len("/service")]
    return host, path


def _matches_profile(profile, host, path):
    host_patterns = [str(h).lower() for h in profile.get("host_patterns", [])]
    if host_patterns and host not in host_patterns:
        return False
    prefixes = profile.get("path_prefixes", [])
    if not prefixes:
        return True
    path_lower = path.lower()
    return any(path_lower.startswith(str(prefix).lower()) for prefix in prefixes)


def _match_mine_profile(service_root):
    host, path = _normalize_service_root(service_root)
    if not host:
        return None

    mines = _load_registry().get("mines", {})
    for _, profile in mines.items():
        if _matches_profile(profile, host, path):
            return profile
    return None


def _normalize_benchmark_profile(name, profiles, fallback_name):
    if not profiles:
        return fallback_name
    if name in profiles:
        return name
    if fallback_name in profiles:
        return fallback_name
    return next(iter(profiles))


def _profile_to_plan(profile_name, profile_data):
    workers = _to_int_list(profile_data.get("workers", []))
    if not workers:
        workers = _to_int_list(DEFAULT_BENCHMARK_PROFILES[DEFAULT_BENCHMARK_FALLBACK_PROFILE]["workers"])
    include_legacy = bool(profile_data.get("include_legacy_baseline", False))
    return {
        "name": profile_name,
        "workers": workers,
        "include_legacy_baseline": include_legacy,
    }


def _normalize_production_profile(name, profiles, workflow):
    workflow = _normalize_workflow(workflow)
    fallback = DEFAULT_PRODUCTION_PROFILE_BY_WORKFLOW[workflow]
    if not profiles:
        return fallback
    if name in profiles:
        return name
    if fallback in profiles:
        return fallback
    return next(iter(profiles))


def _production_profile_to_plan(profile_name, profile_data, workflow):
    workflow = _normalize_workflow(workflow)
    default_workers = DEFAULT_WORKERS_TIER
    if profile_name in DEFAULT_PRODUCTION_PROFILES:
        default_workers = int(DEFAULT_PRODUCTION_PROFILES[profile_name].get("workers", DEFAULT_WORKERS_TIER))

    workers = profile_data.get("workers", default_workers)
    try:
        workers = int(workers)
    except Exception:
        workers = default_workers
    if workers <= 0:
        workers = default_workers

    plan_workflow = _normalize_workflow(profile_data.get("workflow", workflow))
    if plan_workflow != workflow:
        plan_workflow = workflow
    if plan_workflow == PRODUCTION_WORKFLOW_ELT:
        default_pipeline = "parquet_duckdb"
    else:
        default_pipeline = "polars_duckdb"

    return {
        "name": profile_name,
        "workflow": plan_workflow,
        "workers": workers,
        "pipeline": str(profile_data.get("pipeline", default_pipeline)),
        "parallel_profile": str(profile_data.get("parallel_profile", "large_query")),
        "ordered": profile_data.get("ordered", "unordered"),
        "large_query_mode": bool(profile_data.get("large_query_mode", True)),
        "prefetch": profile_data.get("prefetch"),
        "inflight_limit": profile_data.get("inflight_limit"),
    }


def _mine_profile_name_for_workflow(mine_profile, *, size, workflow, fallback_profile):
    workflow = _normalize_workflow(workflow)
    switch = mine_profile.get("production_profile_switch_rows", mine_profile.get("large_query_threshold_rows"))
    try:
        threshold = int(switch) if switch is not None else DEFAULT_PROFILE_SWITCH_ROWS
    except Exception:
        threshold = DEFAULT_PROFILE_SWITCH_ROWS

    small_key = f"production_{workflow}_small_profile"
    large_key = f"production_{workflow}_large_profile"
    small_profile = mine_profile.get(small_key)
    large_profile = mine_profile.get(large_key)

    if not small_profile:
        small_profile = _workers_to_production_profile(workflow, int(mine_profile.get("default_workers", DEFAULT_PARALLEL_WORKERS)))
    if not large_profile:
        large_workers = int(mine_profile.get("workers_above_threshold", mine_profile.get("default_workers", DEFAULT_PARALLEL_WORKERS)))
        large_profile = _workers_to_production_profile(workflow, large_workers)

    if size is None:
        unknown_workers = int(
            mine_profile.get(
                "workers_when_size_unknown",
                mine_profile.get("workers_above_threshold", mine_profile.get("default_workers", DEFAULT_PARALLEL_WORKERS)),
            )
        )
        return _workers_to_production_profile(workflow, unknown_workers)

    if int(size) > threshold:
        return str(large_profile or fallback_profile)
    return str(small_profile or fallback_profile)


def resolve_production_plan(
    service_root,
    size,
    *,
    workflow=PRODUCTION_WORKFLOW_ELT,
    production_profile="auto",
):
    workflow = _normalize_workflow(workflow)
    registry = _load_registry()
    profiles = registry.get("production_profiles", {}) or DEFAULT_PRODUCTION_PROFILES

    if production_profile is not None and str(production_profile).strip().lower() != "auto":
        resolved_name = _normalize_production_profile(str(production_profile), profiles, workflow)
        return _production_profile_to_plan(resolved_name, profiles[resolved_name], workflow)

    fallback_name = DEFAULT_PRODUCTION_PROFILE_BY_WORKFLOW[workflow]
    mine_profile = _match_mine_profile(service_root)
    if mine_profile is None:
        fallback_resolved = _normalize_production_profile(fallback_name, profiles, workflow)
        return _production_profile_to_plan(fallback_resolved, profiles[fallback_resolved], workflow)

    profile_name = _mine_profile_name_for_workflow(
        mine_profile,
        size=size,
        workflow=workflow,
        fallback_profile=fallback_name,
    )
    resolved_name = _normalize_production_profile(profile_name, profiles, workflow)
    return _production_profile_to_plan(resolved_name, profiles[resolved_name], workflow)


def resolve_preferred_workers(service_root, size, fallback_workers):
    mine_profile = _match_mine_profile(service_root)
    if mine_profile is None:
        return fallback_workers
    plan = resolve_production_plan(service_root, size, workflow=PRODUCTION_WORKFLOW_ELT, production_profile="auto")
    try:
        return int(plan["workers"])
    except Exception:
        return fallback_workers


def resolve_benchmark_plan(service_root, size, fallback_profile=DEFAULT_BENCHMARK_FALLBACK_PROFILE):
    registry = _load_registry()
    profiles = registry.get("benchmark_profiles", {}) or DEFAULT_BENCHMARK_PROFILES
    fallback_name = _normalize_benchmark_profile(fallback_profile, profiles, DEFAULT_BENCHMARK_FALLBACK_PROFILE)
    mine_profile = _match_mine_profile(service_root)

    if mine_profile is None:
        return _profile_to_plan(fallback_name, profiles[fallback_name])

    threshold = mine_profile.get("benchmark_switch_threshold_rows")
    if (
        threshold is not None
        and size is not None
        and size <= int(threshold)
        and isinstance(mine_profile.get("benchmark_small_workers"), list)
    ):
        workers = _to_int_list(mine_profile.get("benchmark_small_workers", []))
        if workers:
            return {
                "name": "benchmark_small_workers_override",
                "workers": workers,
                "include_legacy_baseline": bool(mine_profile.get("benchmark_small_include_legacy_baseline", False)),
            }

    profile_name = mine_profile.get("benchmark_small_profile", mine_profile.get("benchmark_profile", fallback_name))
    if threshold is not None and size is not None and size > int(threshold):
        profile_name = mine_profile.get("benchmark_large_profile", profile_name)
    profile_name = _normalize_benchmark_profile(str(profile_name), profiles, fallback_name)
    return _profile_to_plan(profile_name, profiles[profile_name])


def resolve_named_benchmark_profile(profile_name, fallback_profile=DEFAULT_BENCHMARK_FALLBACK_PROFILE):
    registry = _load_registry()
    profiles = registry.get("benchmark_profiles", {}) or DEFAULT_BENCHMARK_PROFILES
    fallback_name = _normalize_benchmark_profile(fallback_profile, profiles, DEFAULT_BENCHMARK_FALLBACK_PROFILE)
    resolved_name = _normalize_benchmark_profile(str(profile_name), profiles, fallback_name)
    return _profile_to_plan(resolved_name, profiles[resolved_name])


def resolve_execution_plan(
    *,
    service_root,
    rows_target,
    explicit_workers,
    benchmark_profile,
    include_legacy_baseline,
):
    workers = _to_int_list(explicit_workers or [])
    if workers:
        return {
            "name": "workers_override",
            "workers": workers,
            "include_legacy_baseline": bool(include_legacy_baseline),
        }

    if benchmark_profile is not None and str(benchmark_profile).strip().lower() != "auto":
        profile_plan = resolve_named_benchmark_profile(str(benchmark_profile), DEFAULT_BENCHMARK_FALLBACK_PROFILE)
    else:
        profile_plan = resolve_benchmark_plan(service_root, rows_target, DEFAULT_BENCHMARK_FALLBACK_PROFILE)

    return {
        "name": profile_plan["name"],
        "workers": list(profile_plan["workers"]),
        "include_legacy_baseline": bool(include_legacy_baseline) and bool(profile_plan["include_legacy_baseline"]),
    }


def resolve_benchmark_phase_plan(
    *,
    service_root,
    rows_target,
    explicit_workers,
    benchmark_profile,
    include_legacy_baseline,
):
    # Backward-compatible alias for older benchmark tooling imports.
    return resolve_execution_plan(
        service_root=service_root,
        rows_target=rows_target,
        explicit_workers=explicit_workers,
        benchmark_profile=benchmark_profile,
        include_legacy_baseline=include_legacy_baseline,
    )



from __future__ import annotations

from tempfile import TemporaryDirectory

from setuptools import Command, setup

ANALYTICS_REQUIRED_METHODS = ("dataframe", "to_parquet", "to_duckdb")


class AnalyticsCheckCommand(Command):
    description = "Run Polars/Parquet/DuckDB compatibility smoke check"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            import duckdb
            import polars as pl
        except Exception as exc:
            raise SystemExit(
                "analyticscheck requires core workflow dependencies (polars, duckdb). "
                "Install package dependencies first (for example: pip install -e .)"
            ) from exc

        from intermine314.query import Query

        missing = [name for name in ANALYTICS_REQUIRED_METHODS if not hasattr(Query, name)]
        if missing:
            raise SystemExit("Missing Query analytics API methods: " + ", ".join(missing))

        with TemporaryDirectory() as tmp:
            parquet_file = f"{tmp}/smoke.parquet"
            pl.DataFrame({"x": [1, 2, 3], "label": ["a", "b", "c"]}).write_parquet(parquet_file)
            con = duckdb.connect(database=":memory:")
            try:
                got = con.execute("select sum(x) from read_parquet(?)", [parquet_file]).fetchone()[0]
            finally:
                con.close()

        if got != 6:
            raise SystemExit(f"Unexpected DuckDB/Parquet result: {got}")

        print("analyticscheck: ok")


setup(cmdclass={"analyticscheck": AnalyticsCheckCommand})



from __future__ import annotations

from urllib.parse import urlparse, urlunparse


SERVICE_ROOT_KEYS = (
    "webServiceRoot",
    "serviceRoot",
    "serviceUrl",
    "service_url",
    "service",
    "url",
)


def normalize_service_root(root):
    text = str(root or "").strip()
    if not text:
        raise ValueError("A service root URL is required")
    if "://" not in text:
        text = "http://" + text
    parsed = urlparse(text)
    path = (parsed.path or "").rstrip("/")
    if not path.endswith("/service"):
        path = path + "/service" if path else "/service"
    normalized = parsed._replace(path=path, params="", query="", fragment="")
    return urlunparse(normalized)


def service_root_from_payload(payload, *, normalize=False):
    for key in SERVICE_ROOT_KEYS:
        value = payload.get(key) if isinstance(payload, dict) else None
        if value:
            return normalize_service_root(value) if normalize else value
    raise KeyError("Could not resolve service root from registry payload")



from __future__ import annotations


def require_int(name, value):
    if not isinstance(value, int) or isinstance(value, bool):
        raise TypeError(f"{name} must be an integer")
    return value


def require_positive_int(name, value):
    ivalue = require_int(name, value)
    if ivalue <= 0:
        raise ValueError(f"{name} must be a positive integer")
    return ivalue


def require_non_negative_int(name, value):
    ivalue = require_int(name, value)
    if ivalue < 0:
        raise ValueError(f"{name} must be a non-negative integer")
    return ivalue


def resolve_parallel_strategy(pagination, start, size, *, valid_parallel_pagination, keyset_auto_min_size):
    strategy = str(pagination).lower()
    if strategy not in valid_parallel_pagination:
        choices = ", ".join(sorted(valid_parallel_pagination))
        raise ValueError("pagination must be one of: %s" % (choices,))

    if strategy == "offset":
        return "offset"
    if strategy == "keyset":
        return "keyset" if start == 0 else "offset"

    if start != 0:
        return "offset"
    if size is None:
        return "keyset"
    if size >= keyset_auto_min_size:
        return "keyset"
    return "offset"


def normalize_order_mode(ordered, *, default_order_mode, valid_order_modes):
    if ordered is None:
        return default_order_mode
    if isinstance(ordered, bool):
        return "ordered" if ordered else "unordered"
    mode = str(ordered).strip().lower()
    if mode not in valid_order_modes:
        choices = ", ".join(sorted(valid_order_modes))
        raise ValueError("ordered must be a bool or one of: %s" % (choices,))
    return mode


def apply_parallel_profile(profile, ordered, large_query_mode, *, default_profile, valid_parallel_profiles):
    if profile is None:
        profile = default_profile
    profile_name = str(profile).strip().lower()
    if profile_name not in valid_parallel_profiles:
        choices = ", ".join(sorted(valid_parallel_profiles))
        raise ValueError("profile must be one of: %s" % (choices,))

    effective_ordered = ordered
    effective_large = bool(large_query_mode)
    if profile_name == "large_query":
        effective_large = True
        if effective_ordered is None:
            effective_ordered = "window"
    elif profile_name == "unordered":
        if effective_ordered is None:
            effective_ordered = False
    elif profile_name == "mostly_ordered":
        if effective_ordered is None:
            effective_ordered = "window"
    else:
        if effective_ordered is None:
            effective_ordered = True

    return profile_name, effective_ordered, effective_large


def resolve_prefetch(prefetch, *, max_workers, large_query_mode, default_parallel_prefetch):
    if prefetch is None:
        if default_parallel_prefetch is not None:
            prefetch = default_parallel_prefetch
        else:
            prefetch = max_workers * 2 if large_query_mode else max_workers
    return require_positive_int("prefetch", prefetch)


def resolve_inflight_limit(inflight_limit, *, prefetch, default_parallel_inflight_limit):
    if inflight_limit is None:
        inflight_limit = (
            default_parallel_inflight_limit if default_parallel_inflight_limit is not None else prefetch
        )
    return require_positive_int("inflight_limit", inflight_limit)
